#! /usr/bin/perl -w
# logging client for Envisalink
#
# Version: 1.0
# 
# code by prozzerg
#  

use Mojo::IOLoop;
use DateTime::Format::Strptime;
use POSIX;
use LWP::Simple;
use JSON qw( decode_json );
use feature qw(switch);
use feature     qw< unicode_strings >;
use utf8;

# disable warnings
no warnings;

# MOJO_IOLOOP_DEBUG=1
           
# information about the Envisalink module
my $password = "ENVISALINK_PASS";
my $host = 'ENVISALINK_IP';
my $port = 4025;

# toggle debug log
my $debug = 0;


# then we got domoticz information
my $domo_host = "DOMOTICZ_IP";
my $domo_port = 8080;

# we also need some zone->idx mapping. lets use an hash table for that
my %domo_mapping = (
                  '001' => 168, 
                  '002' => 166, 
                  '003' => 167,
                  '004' => 169,
                  '005' => 171,
                  '006' => 170,
                  '007' => 176,
                  '008' => 177,
                  '009' => 172,
                  '010' => 173,
                  '011' => 174,
                  '012' => 175
                );
				
# we also need some partition->idx mapping. lets use an hash table for that
# This reflects two of my partition in my DSC dystem (1 and 2).
# You must specify each partition twice, for partition no 1: 1 and 10, for paartition no 2: 2 and 20 and so on
# I know it's dirty but I have been strugling a while with it and didn't fix it. So leave like it is or fix and share back to me if you can :)
my %domo_partition_mapping = (
                  '1' 	=> 181, 
                  '10' 	=> 181, 
                  '2' 	=> 182,
                  '20' 	=> 182
                );

				
# define subroutines
sub handle_data($);
sub validate_data($);
sub extract_data($);
sub get_partuser($);
sub create_checksum($);
sub create_date($);
sub display_debug($);
sub display_regular($);
sub log_zone;
sub log_partition;
sub get_domo_status($);
sub set_domo_status;
my $loggedin = 0;

# Connect to envisalink
my $id = Mojo::IOLoop->client({address => $host, port => $port} => sub {
  my ($loop, $err, $stream) = @_;
  $stream->timeout(300);
  $stream->on(read => sub {
    my ($stream, $bytes) = @_;

    # Process input
    # split it
    my @lines = split('\n', $bytes);
    foreach my $line (@lines)
    {
    	my $writedata = handle_data($line);
    	if ($writedata ne "") {
    	  $stream->write("$writedata\r\n");
      }
    }
  });
});

# do stuff each minute
  Mojo::IOLoop->recurring(60 => sub {
	my $loop = shift;
	if ($loggedin)
	{
		my $stream = Mojo::IOLoop->stream($id);
		display_regular("Recurring status request...");
		# pull report
		my $command = "001" . create_checksum("001");
		$stream->write("$command\r\n");
	}
});

# Start event loop if necessary
Mojo::IOLoop->start unless Mojo::IOLoop->is_running;

# -----------------------------------------------------------------------
# all the subroutines
sub handle_data($)
{
	local $inputline = $_[0];
	local $output = "";

	chomp($inputline);
	
	# validate input
	if (!(validate_data($inputline)))
	{
		display_regular("Data was invalid! Input -> $inputline");
		return $output;
	}

	# split it
	my ($command,$data) = extract_data($inputline);
	
	# handle the data, start with the login
  if ($command eq "505")
	{
		# login function
		if ($data eq "0")
		{
			display_regular("login FAILED");
		} elsif ($data eq "1")
		{
			display_regular("login OK");
			$loggedin = 1;
			$output = "001";
		} elsif ($data eq "2")
		{
			display_regular("login Time.out");
		} elsif ($data eq "3")
		{
			display_regular("Password Request, sending...");
			$output = "005$password";
		}
	} 
	
	# lets do a "switch/case" for the rest
	given ($command) {
		when("500") { display_regular("ACK for command: $data"); } # ACK for last command
		when("505") { } # login, handled above. added to remove "unknown command" output
		when("510") { display_regular("Keypad LED State: $data"); } # Keypad LED State, partition 1
		when("511") { display_regular("Keypad LED FLASH State: $data"); } # Keypad LED FLASH state, partition 1
		when("550") { local $tempdate = create_date($data); display_regular("Got date: $tempdate"); } # Date broadcast
		when("601") { display_regular("Zone Alarm: $data"); log_zone($command, $data); } # Zone Alarm
		when("602") { display_regular("Zone Alarm Restore: $data"); log_zone($command, $data); } # Zone Alarm Restore
		when("603") { display_regular("Zone Tamper: $data"); log_zone($command, $data); } # Zone Tamper
		when("604") { display_regular("Zone Tamper Restore: $data"); log_zone($command, $data); } # Zone Tamper Restore
		when("605") { display_regular("Zone Fault: $data"); log_zone($command, $data); } # Zone Fault
		when("606") { display_regular("Zone Fault Restore: $data"); log_zone($command, $data); } # Zone Fault Restore
		when("609") { display_regular("Zone Open: $data"); log_zone($command, $data); } # Zone Open
		when("610") { display_regular("Zone Restore: $data"); log_zone($command, $data); } # Zone Restore
		when("650") { display_regular("Partition Ready: $data"); } # Partition Ready
		when("651") { display_regular("Partition Not Ready: $data"); } # Partition Not Ready
		when("652") { display_regular("Partition Armed: $data"); log_partition($command, $data); } # Partition Armed
		when("653") { display_regular("Partition Ready \- Force Arming Enabled: $data"); } # artition Ready  Force Arming
		when("654") { display_regular("Partition In Alarm: $data"); } # Partition In Alarm
		when("655") { display_regular("Partition Disarmed: $data"); log_partition($command, $data); } # Partition Disarmed
		when("656") { display_regular("Partition Exit Delay in Progress: $data"); } # Exit Delay in Progress
		when("657") { display_regular("Partition Entry Delay in Progress: $data"); } # Entry Delay in Progress
		when("658") { display_regular("Partition Keypad Lock-out: $data"); } # Keypad Lock-out
		when("659") { display_regular("Partition Failed to Arm: $data"); } # Failed to Arm
		when("660") { display_regular("Partition PGM Output is in Progress: $data"); } # PGM Output is in Progress
		when("663") { display_regular("Partition Chime Enabled: $data"); } # Chime Enabled
		when("664") { display_regular("Partition Chime Disabled: $data"); } # Chime Disabled
		when("670") { display_regular("Partition Invalid Access Code: $data"); } # Invalid Access Code
		when("671") { display_regular("Partition Function Not Available: $data"); } # Function Not Available
		when("672") { display_regular("Partition Failure to Arm: $data"); } # Failure to Arm
		when("673") { display_regular("Partition is Busy: $data"); } # Partition is Busy
		when("674") { display_regular("Partition System Arming in Progress: $data"); } # System Arming in Progress
		when("700") { local ($partition, $user) = get_partuser($data); display_regular("Partition User Closing: $partition by User: $user"); } # User Closing
		when("701") { display_regular("Partition Special Closing: $data"); } # Special Closing
		when("702") { display_regular("Partition Partial Closing: $data"); } # Partial Closing
		when("750") { local ($partition, $user) = get_partuser($data); display_regular("Partition User Opening: $partition by User: $user "); } # User Opening
		when("841") { display_regular("Trouble Led off for Partition: $data"); } # Trouble Led OFF
		default { display_regular("Unknown command: $command with data: $data"); } # Catching the rest
	};

  # when("606") {  } # 
	
	# add checksum to the output if it isnt empty
	if ($output ne "")
	{
		$output = $output . create_checksum($output);
	}
	
	return $output;
}

sub validate_data($)
{
	my $inputline = $_[0];
	my $dec_total = 0;
	my $command = substr $inputline, 0, length($inputline)-3;
	my $checksum = substr $inputline, length($inputline)-3, 2;
	display_debug("command: $command  checksum: $checksum");
	
	#calculate checksum
	local $newchecksum = create_checksum($command);
	display_debug("new checksum: $newchecksum");
	
	if ($newchecksum eq $checksum)
	{
		return 1;
	} else
	{
		return 0;
	}
}

sub extract_data($)
{
	my $inputline = $_[0];
	my $cmd = substr $inputline, 0, 3;
	my $data = substr $inputline, 3, length($inputline)-6;
	return($cmd, $data);
}

sub get_partuser($)
{
	my $data = $_[0];
	my $partition = substr $data, 0, 1;
	my $user = substr $data, 1;
	return($partition, $user);
}

sub create_checksum($)
{
  my $command = $_[0];
  my $dec_total = 0;
  for ($i=0; $i < length($command); $i++)
	{
		$dec_total = $dec_total + ord(substr $command, $i, 1);
	}
	my $checksum = uc(sprintf("%x", ($dec_total % 256)));
	return $checksum;
}

sub create_date($)
{
  my $data = $_[0];
  my $date = "";
  
  # date format is 1832 07 07 13 
  my $parser = DateTime::Format::Strptime->new(
   pattern => '%H%M%m%d%y',
   on_error => 'croak',
  );
  my $dt = $parser->parse_datetime($data);
  $date = $dt->strftime('%F %R');
	return $date;
}

sub display_debug($)
{
	my $msg = $_[0];
	if ($debug)
	{
		display_regular("DEBUG - $msg");
	}
}

sub display_regular($)
{
	my $msg = $_[0];
	my $timestamp = POSIX::strftime("%Y-%m-%d %H:%M:%S", localtime);
	print "$timestamp - $msg\n";
}

sub log_zone
{
	my $status = $_[0];
	my $zone = $_[1];
	
	# this is kinda ugly, but we need to strip the partition part of the zone if it exists (stupid dscserver!)
	if (length($zone) eq 4) {
		$zone = substr $zone, 1;
	}
	
	my $newstatus = "";
	given ($status) {
		when("601") { $newstatus = "alarm" }
		when("602") { $newstatus = "alarm restore" }
		when("603") { $newstatus = "tamper" }
		when("604") { $newstatus = "tamper restore" }
		when("605") { $newstatus = "fault" }
		when("606") { $newstatus = "fault restore" }
		when("609") { $newstatus = "open" }
		when("610") { $newstatus = "restored" }
	};
	
  # send to domoticz
  if (exists $domo_mapping{$zone} ) {
  	display_debug("we got a mapping for zone '$zone' -> $domo_mapping{$zone}, trying to send to domoticz");
  	my $domostatus = "On";
  	if ($status eq "602" || $status eq "604" || $status eq "606" || $status eq "610") {
  		$domostatus = "Off";
  	}
  	# send status to domoticz
  	# but first, check it
  	if (get_domo_status($domo_mapping{$zone}) ne $domostatus) {
		if ($domostatus ne "Off") {
			display_debug("sending '$domostatus' for idx $domo_mapping{$zone}");
 	 		set_domo_status("http://$domo_host:$domo_port/json.htm?type=command&param=switchlight&idx=$domo_mapping{$zone}&switchcmd=$domostatus");
		}
    } else {
    	display_debug("idx $domo_mapping{$zone} already got status '$domostatus'");
    }
  } else {
  	display_debug("no mapping for zone '$zone'");
  }
}

sub log_partition
{
	my $status = $_[0];
	my $partition = $_[1];
	
	# this is kinda ugly, but we need to strip the partition part of the zone if it exists (stupid dscserver!)
	if (length($partition) eq 4) {
		$partition = substr $partition, 1;
	}
	
	#my $newstatus = "";
	#given ($status) {
	#	when("601") { $newstatus = "alarm" }
	#	when("602") { $newstatus = "alarm restore" }
	#	when("603") { $newstatus = "tamper" }
	#	when("604") { $newstatus = "tamper restore" }
	#	when("605") { $newstatus = "fault" }
	#	when("606") { $newstatus = "fault restore" }
	#	when("609") { $newstatus = "open" }
	#	when("610") { $newstatus = "restored" }
	#};
	
	# here we can do something with the data. placeholder for now
	# send to domoticz
	if (exists $domo_partition_mapping{$partition} ) {
		display_debug("we got a mapping for partition '$partition' -> $domo_partition_mapping{$partition}, trying to send to domoticz");
		my $domostatus;
		my $domostatus_nvalue;
		
		if ($status eq "652") { # Partition armed
			$domostatus = "On";
			$domostatus_nvalue = "0";
		} elsif ($status eq "655") { # Partition disarmed
			$domostatus = "Off";
			$domostatus_nvalue = "1";
		}
		# send status to domoticz
		# but first, check it
		if (get_domo_status($domo_partition_mapping	{$partition}) ne $domostatus) {
			display_debug("sending '$domostatus' for idx $domo_partition_mapping{$partition}");
			#set_domo_status("http://$domo_host:$domo_port/json.htm?type=command&param=udevice&idx=$domo_partition_mapping{$partition}&nvalue=$domostatus_nvalue");
			set_domo_status("http://$domo_host:$domo_port/json.htm?type=command&param=switchlight&idx=$domo_partition_mapping{$partition}&switchcmd=$domostatus");
		} else {
			display_debug("idx $domo_partition_mapping{$partition} already got status '$domostatus'");
		}
	} else {
		display_debug("no mapping for zone '$partition'");
	}
	
}

sub get_domo_status($)
{
	my $idx = $_[0];
	
	my $domostatusurl = "http://$domo_host:$domo_port/json.htm?type=devices&filter=all&used=true&order=Name&rid=$idx";
	
	my $json = get( $domostatusurl );
  die "Could not get $domostatusurl!" unless defined $json;
  
  my $decoded = JSON->new->utf8(0)->decode( $json );
  
  my @results = @{ $decoded->{'result'} };
  my $status = $results[0]->{"Data"};
  
  display_debug("domostatus for idx $idx -> '$status'");
	
	return $status;
}

sub set_domo_status
{
	my $domourl = $_[0];
	my $response = get( $domourl );
  die "Could not get $domourl" unless defined $response;	
}
